#!/usr/bin/perl -w
#############################################################################
# PROGRAM
#     dst
#
# SYNOPSIS
#     Used to determine various country DST functions
#
# USAGE
#     dst [-c (uk/us)] [-b] [-d] [-h] [-y year (4 digits)] 
#
# PARAMETERS
#     None
#
# OPTIONS
#     -c [uk/us]	= country ID
#     -b		= returns 0 or 1 (bit) to indicate whether DST is active or not
#     -d		= returns a nice sentence to indicate whether DST is active or not 
#     -h		= usage screen
#     -y [year]		= returns the dates the specified country is under DST for specified year
#			  if no year is specified, it defaults to the current year
#
# MODULES/FILES REQUIRED
#     File::Basename
#     Time::Local
#     Date::Calc
#
# PROGRAMS REQUIRED
#     None
#
# AUTHOR
#     Steve Dickinson
#
# DEVELOPMENT DATES
#     03.24.08
#
# RELEASE DATE
#     03.24.08
#
# COMMENTS
#     None
#
# REVISIONS
#     None
#############################################################################
# INTERRUPT SIGNALS
#############################################################################
$SIG{'INT' } = 'clean_up'; $SIG{'HUP' } = 'clean_up';
$SIG{'QUIT'} = 'clean_up'; $SIG{'TRAP'} = 'clean_up';
$SIG{'ABRT'} = 'clean_up'; $SIG{'STOP'} = 'clean_up';
$SIG{'TSTP'} = 'clean_up';

#############################################################################
# INCLUDE
#############################################################################
use File::Basename;
use Time::Local;
use Date::Calc qw(:all);

use strict;

#############################################################################
# USAGE
#############################################################################
my $prog = basename($0,"");

sub usage {
    print "\nusage: $prog [-c uk/us] [-b] [-d] [-h] [-y year (4 digits)]\n\n";
    print "Options:\n";
    print "  -c [uk/us]                 country ID (required)\n";
    print "  -b                         returns 0 or 1 to indicate whether DST is not active or is active respectively\n";
    print "  -d                         returns a nice sentence to indicate whether DST is active or not\n";
    print "  -h                         usage information\n";
    print "  -y [year] (4 digits)       returns the dates the specified country is under DST for specified year\n";
    print "                             if no year is specified, it defaults to the current year\n";
    print "\n";
}

#############################################################################
# PROCESS ARGUMENTS
#############################################################################
my $opt = "";
my @current_time;
our $arg_bit = "false";
our $arg_cid = "";
our $arg_dst = "false";
our $arg_dates = "false";
our $arg_year = "";
our $current_year = "";
@current_time = localtime(time);
$current_year = $current_time[5] + 1900; 
$arg_year = $current_year; 
our $check_cid = "false";
our $check_year = "false";

for (my $argc = 0; $argc <= $#ARGV; $argc++) {
    if ($ARGV[$argc]) {
        $opt =  $ARGV[$argc];
        $opt =~ s/--//; # Get rid of 2 dashes
        $opt =~ s/-//; # Get rid of 1 dash
        $opt =  substr($opt,0,1); # cut the first char
    }
	
    if ($opt eq 'c') {
	$arg_cid = $ARGV[++$argc];
	if ($arg_cid) {
	    $arg_cid = lc($arg_cid);
  	    $check_cid = check_cid($arg_cid);
	    if ($check_cid eq "false") {
	        usage();
 	        exit 1;
	    }
        } else {
	    usage();
	    exit 1;
	}
    } elsif ($opt eq 'b') {
	$arg_bit = "true"; 
    } elsif ($opt eq 'd') {
	$arg_dst = "true"; 
    } elsif ($opt eq 'h') {
	usage();
	exit 0;
    } elsif ($opt eq 'y') {
	$arg_dates = "true";
	$arg_year = $ARGV[++$argc];
	if ($arg_year) {
  	    $check_year = check_year($arg_year);
	    if ($check_year eq "false") {
	        usage();
 	        exit 1;
	    }
        } else {
	    $arg_year = $current_year;
	}
    } else {
	usage();
	exit 1;
    }
}

if (!$arg_cid || ($arg_bit eq "false" && $arg_dst eq "false" && $arg_dates eq "false")) {
    usage();
    exit 1;
}

#============================================================================
# CHECK YEAR SYNTAX
#============================================================================
sub check_year {
    my $status = "false";
    my $year = shift;
    if ($year =~ /^[0-9]{4}$/) {
        $status = "true";
    }
 
    return ($status);    
}

#============================================================================
# CHECK COUNTRY ID SYNTAX
#============================================================================
sub check_cid {
    my $status = "false";
    my $arg_cid = shift;

    if ($arg_cid =~ /^uk$|^us$/) {
        $status = "true";
    }
 
    return ($status);    
}

#############################################################################
# GLOBAL VARIABLES / SETTINGS
#############################################################################
umask(0);

use constant DAYS_OF_THE_WEEK => ["Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"]; 
use constant MONTHS => ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]; 

our ($start_month, $end_month, $start_hour, $end_hour);

if ($arg_cid =~ /^uk$/) {
    $start_month = 3;
    $end_month = 10;
    $start_hour = 1;
    $end_hour = 1;
}

if ($arg_cid =~ /^us$/) {
    $start_month = 3;
    $end_month = 11;
    $start_hour = 2;
    $end_hour = 2;
}

#############################################################################
# MAIN
#############################################################################
main();
sub main {
    if ($arg_bit eq "true") {
        get_dst($arg_year);
    } elsif ($arg_dst eq "true") {
        get_dst($arg_year);
    } elsif ($arg_dates eq "true") {
	get_dates($arg_year);
    } 
    exit(0);
}

#============================================================================
# DETERMINE IF COUNTRY IS CURRENTLY UNDER DST
#============================================================================
sub get_dst {
    my (@UTC_time, $UTC_year);
    my ($UTC_current_seconds, $UTC_sunday_of_start_month_seconds, $UTC_sunday_of_end_month_seconds);
    my ($sunday_of_start_month, $sunday_of_end_month, $country, $dst);
    my $year = (shift) - 1900;

    @UTC_time = gmtime(time);
    $UTC_year = $UTC_time[5];

    $UTC_current_seconds = timegm(@UTC_time);

    if ($arg_cid eq "uk") {
        $sunday_of_start_month = get_last_sunday($start_month, $year+1900);
        $sunday_of_end_month = get_last_sunday($end_month, $year+1900);
	$country = "United Kingdom";
	$dst = "British Summer Time (BST)";
    }

    if ($arg_cid eq "us") {
        $sunday_of_start_month = get_second_sunday($start_month, $year+1900);
        $sunday_of_end_month = get_first_sunday($end_month, $year+1900);
	$country = "United States";
	$dst = "Daylight Savings Time (DST)";
    }

    $UTC_sunday_of_start_month_seconds=timegm(0,0,$start_hour,$sunday_of_start_month,$start_month-1,$UTC_year);
    $UTC_sunday_of_end_month_seconds=timegm(0,0,$end_hour,$sunday_of_end_month,$end_month-1,$UTC_year);

    if (($UTC_current_seconds > $UTC_sunday_of_start_month_seconds) && ($UTC_current_seconds < $UTC_sunday_of_end_month_seconds)) {
	if ($arg_bit eq "true") {
	    print "1\n";
	} else {
            print "Currently the $country IS under ${dst}.\n";
	}
    } else {
	if ($arg_bit eq "true") {
	    print "0\n";
	} else {
        print "Currently the $country IS NOT under ${dst}.\n";
	}
    }
}

#============================================================================
# GET DATES OF DST FOR CURRENT YEAR
#============================================================================
sub get_dates {
    my (@UTC_sunday_of_start_month_time, @UTC_sunday_of_end_month_time);
    my ($start_month_hour, $end_month_hour, $start_month_meridiem, $end_month_meridiem);
    my ($sunday_of_start_month, $sunday_of_end_month, $country, $dst);
    my $year = (shift) - 1900;

    if ($arg_cid eq "uk") {
	$country = "UK";
	$dst = "British Summer Time (BST)";
        $sunday_of_start_month = get_last_sunday($start_month, $year+1900);
        $sunday_of_end_month = get_last_sunday($end_month, $year+1900);
    }

    if ($arg_cid eq "us") {
	$country = "US";
	$dst = "Daylight Savings Time (DST)";
	if ($year < 107) {
	    print "The standards for $dst in the $country were changed in 2007.\n";
	    print "Therefore, this program does not retrieve values prior to 2007.\n";
	    exit;
	}
        $sunday_of_start_month = get_second_sunday($start_month, $year+1900);
        $sunday_of_end_month = get_first_sunday($end_month, $year+1900);
    }

    # get epoch time of last day of March
    # convert epoch time to UTC list
    @UTC_sunday_of_start_month_time = gmtime(timegm(0,0,$start_hour,$sunday_of_start_month,$start_month-1,$year));
    @UTC_sunday_of_end_month_time = gmtime(timegm(0,0,$end_hour,$sunday_of_end_month,$end_month-1,$year));

    $start_month_hour = $UTC_sunday_of_start_month_time[2];
    if ($start_month_hour >= 0 && $start_month_hour < 12) {
	$start_month_meridiem = "am";
    } else {
	$start_month_meridiem = "pm";
    }
    $end_month_hour = $UTC_sunday_of_end_month_time[2];
    if ($end_month_hour >= 0 && $end_month_hour < 12) {
	$end_month_meridiem = "am";
    } else {
	$end_month_meridiem = "pm";
    }

    $year += 1900;
  
    print "$dst takes affect in the $country during $year as follows:\n\n";
    print "Start: " . $UTC_sunday_of_start_month_time[2] . $start_month_meridiem . " " . DAYS_OF_THE_WEEK->[$UTC_sunday_of_start_month_time[6]] . " " . MONTHS->[$UTC_sunday_of_start_month_time[4]] . " " .  $UTC_sunday_of_start_month_time[3] . ", $year\n";
    print "  End: " . $UTC_sunday_of_end_month_time[2] . $end_month_meridiem . " " . DAYS_OF_THE_WEEK->[$UTC_sunday_of_end_month_time[6]] . " " . MONTHS->[$UTC_sunday_of_end_month_time[4]] . " " .  $UTC_sunday_of_end_month_time[3] . ", $year\n"; 
}

#============================================================================
# GET THE DAY NUMBER OF THE LAST SUNDAY OF THE SPECIFIED MONTH
#============================================================================
sub get_first_sunday {
    my ($year, $month, $day, @time);
    $month = shift;
    $year = shift;

    my $days = Days_in_Month($year, $month);

    for ($day=1; $day<=$days; $day++) {
	@time = gmtime(timegm(0,0,1,$day,$month-1,$year-1900));
	if ($time[6] == 0) {
	    last;
	}
    }

    return ($day);
}

#============================================================================
# GET THE DAY NUMBER OF THE LAST SUNDAY OF THE SPECIFIED MONTH
#============================================================================
sub get_second_sunday {
    my ($year, $month, $day, @time);
    $month = shift;
    $year = shift;

    my $sunday_count = 0;
    my $days = Days_in_Month($year, $month);

    for ($day=1; $day<=$days; $day++) {
	@time = gmtime(timegm(0,0,1,$day,$month-1,$year-1900));
	if ($time[6] == 0) {
	    $sunday_count++;
 	    if ($sunday_count == 2) {
	  	last;
	    }
	}
    }

    return ($day);
}

#============================================================================
# GET THE DAY NUMBER OF THE LAST SUNDAY OF THE SPECIFIED MONTH
#============================================================================
sub get_last_sunday {
    my ($year, $month, $day, @time);
    $month = shift;
    $year = shift;

    my $days = Days_in_Month($year, $month);

    for ($day=$days; $day>0; $day--) {
	@time = gmtime(timegm(0,0,1,$day,$month-1,$year-1900));
	if ($time[6] == 0) {
	    last;
	}
    }

    return ($day);
}

#############################################################################
# CLEAN UP
#############################################################################
sub clean_up {
    exit;
}
